!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!
! Module      : omprep_mod
! Authors     : Tim Harries (MSSL)
!               Jon Rainger (MSSL)
!               Phil Williams (MSSL)
!               Andrew King (MSSL)
!               Chris Brindle (MSSL)
!               Vladimir Yershov (MSSL)
! Description : This tasks interacts directly with the ODF files, cloning the input and 
!               modifying it for subsequent processing  by the pipeline tasks. 
!
! Comments    : SCIWIN convention not stable yet.
!               FAST Mode SAMPTIME and BFAST_ID keywords could be used in place of BINAX1, BINAX2.
!               Origin of RA_PNT, DEC_PNT and ROLL_PNT unknown. Possibly could be taken from SUM file.
!               Value for keyword ORIGIN unknown.
!
! $Id         : module.f90,v 1.4 1997/12/18 16:01:10 gvacanti Exp $
!
! History : Created  : January 98 - TJH
!           Modified : 27 April 98 - Removed datestamp
!                      12 March 99 - Spring clean of code.
!                      20 April 99 - Change parameter names to conform to new convention.
!                      18 May - Remove dependancies on filename.
!                             - Remove odfdir parameter and call to open OAL.
!                      21 July - Writes sample time and bfast id to event list.
!                      06 Jan - Additional RAW extension added.rENG
!                      17.05.2000 - Removed hardwiring of dimension of 10 for maximum number of 
!                      guide stars although this was correct. WE saw some data with a maximum of 16.
!                      The code now works out the maximum dimension size from the data and uses this
!                      Should work in all cases! 
!                      23.05.2000 AMK - Implemented calls to OAL_getodInfo and OAL_getAttitude
!                      to get telescope attitude. Remomoved hardwiring RA,DEC and PA and replaced
!                      with values obtaind from the above calls 
!                      21.06.2000 AK - Implemented the call to OAL_exposureInfo in order that the
!                      start time, end time and exposure length could be determined.
!                      The call to getExposure has been removed.
!                      20.07.2000 AMK - Added a command line parameter slewset which indicates whether
!                      we are processing slew data or true imaging data. The logical flag slewFlag is set
!                      accordingly. There are two switches in the code that depend upon this slewFlag.
!                      The first switch just sets binBPE to TRUE if slewFlag is TRUE otherwise the code works
!                      out as before how to set binBPE.
!                      The second switch will skip the attitude code if slewFlag is TRUE.
! 21/07/2000 AMK - Added additional code to deal with the case of a missing tracking 
! history file.
! OMCHAIN has been modified so that when this situation arises the filename dummyTHX.FIT is passed
! in. The code now tests for this situation and set the flag dummyTHXFlag to TRUE. The 
! ZERODRIFT option is then implemented.
! This device enables us to get around this problem and requiring the user to have a dummy file
! present.
! 21/07/2000 AMK - Added the above creation of a dummy tmp_tracking for the slew optin.
! 24/07/2000 AMK - Added the subroutine convertExposureStringToInteger. This converts the
! string holding the exposure number to an integer. This is in turn passed to getstartendexp.  
! 24/08/2000 AMK - Added the keyword BLANK to the output image file. This is required later
! in the chain by OMFLATFIELD in order to set the quality array.
! 22/09/2000 AMK - implemented checking of the raw image in order to detect negative pixel counts.
! These counts are reset to zero and a warning issued. This is short term fix.
! 02/10/2000 AMK changed the call t getStartEndExp which now returns the exposure filter.
! The code that obtains he filter from the PEH file is now redundant.
! 18/10/2000 - Filter string initialised to " "
! 1.11.2000 - The filter string is now checked to ensure it is a valid filter- if not then the
! program aborts.
! 14.11.2000 - DETNAM set to REDUNDANT instead of PRIME if chain=1. 
! 06.03.2001 - routine getexposurestartrow revised
! 07.03.2001 - subroutine CheckImageSizeOkay added to ensure that the x and y image dimensions
! agree with those computed from the fits file attributes BINBPE, WINDOWDX, WINDOWDY, BINAX1 
! and BINAX2
! 03.05.2001 VNY - instrumentMode integer variabe added (IMAGING, DATA_FAST, etc.),
! necessary for the evselect procedure
! 11.05.2001 VNY - PHA column added  to the output file
! 11.05.2001 VNY -  TLMIN, TLMAX parameters have been changed in order to correct
!            image construction by evselect
! 26.05.2001 VNY - when an incorrect filter name is found  stops  only in the Imaging Mode; 
!            in the Fast Mode makes only a warning message 
! 27.06.2001 VNY - in case the filter attribute is not found in the THX-file (fast mode) 
!            it is taken from the FAE-file
! 11.07.2001 VNY - checking negative values in the input FAST event list and reducing the output 
!            list if any negative values were found
! 31.07.2001 VNY - fixing float point exception (MATRIX array changed from (3,3) to (0:2,0:2))
!            25.09.2001 VNY - TIME column in the output event list file was wrong.
! 25.09.2001 CB  - BINBPE parameter obtained from the image header - call to GetStartEndExposure
!           06.11.2001 VNY - 1-pixel difference in the calculation of DETSEC parameters has been rectified
! 05.12.2001 CB - Check added to see if exposure time picked up from raw-image header-file is zero-
!            if so computes it from nph file- if then zero program aborts
! 20.02.2002 CB - Added call to OAL_addCommonAttributes to propagate new keywords. Also added the 
!            keyword "EXPIDSTR"
! 12.09.2002 VNY - minor corrections (in order to include the frametime_mod.f90 module)
! 16.06.2003 VNY - The subroutine GetExposureIdentifier can crash if the input tracking 
!             history filename is DUMMYTHX.FIT
!            To prevent this, the exposure identifier is picked up from the PEH-file 
!            for the case of DUMMYTHX
! 16.10.2003 VNY - corrections to work with the grism images (the modeset=4)
! 23.10.2003 VNY - binax1 and binax2 are computed as in the previous version (2.32). This is 
!            not correct, but temporarily it is possible to leave it without changes
! 29.06.2004 VNY - a bug is fixed that was producing wrong window sizes in the DETSEC keyword (SSC-SPR-3345)
! 16.02.2005 VNY - trying to resolve the Eng-4 data problem
! 09.06.2005 VNY - some warning messages have the second part of the message empty, which causes a very 
! long strings of nonsense to appear in these message 
! 12.09.2005 VNY - trying to recover the data corresponding to negative velues in the input file
! 10.11.2005 VNY - the problem with oscillating sign of FTFINE is fixed 
! 09.09.2007 VNY - extending the range of the DATAMODE parameter: for the value 5 the task will not
!                  use the SUM.SAS-file assuming the simulated FAST data. 
!                  The exposure identifier must be extracted from the header keyword EXP_ID rather
!                  then from the file name (this is yet to be fixed).
! 2009-05-29 VNY  Protecting against wrong binning factors in the WDX table (SSC-SPR-6541)
! 2009-06-01 VNY  Adding a warning message for the cases when a binning factor in the WDX table
!                 was found wrong (SSC-SPR-6541)
! 2011-03-22 VNY using new parameter rawattitude; fixed a problem of the 
! interface between F90 and C++ which was causing a run-time error
! 2011-03-23 VNY the previous fix did not work; so fixing again the problem with the 
! run-time exception error 
! 2011-03-24 VNY Correcting the task's behaviour for the modeset=3, which should ignore
! the parameter rawattitude
! 2012-07-19 VNY checked the initialisation of variables dummyTHXFlag;
! propagated the keywords BFASTID and SAMPTIME in the output file header 
! 
! 2013-07-13 VNy protecting against the possibility of calculating a wrong binning
! factor for full-frame high-resolution images
! 
!***********************************************************************************************
module omprep_mod

  use param
  use oal
  use caloaldefs
  use dal

  use sasdal
  use errorHandling
  use getstartendexp_mod

  use frametime_mod 
  !
  !-----------------------------------------------------------------------
  !
  use types
  !
  !----------------------------------------------------------------------
  !
  implicit none

  private

  public omprep
 
contains

  !---------------------------------------------------------------------
  subroutine omprep
    !---------------------------------------------------------------------
    type(DataSetT) :: inDataSetH     ! Input dataset handle.
    type(DataSetT) :: pehDataSetH    ! Periodic Housekeeping dataset handle.
    type(DataSetT) :: wdxDataSetH    ! Priority Window Data dataset handle.
    type(DataSetT) :: outDataSetH    ! Output dataset handle.
    type(ArrayT) :: inArrayH         ! Input array handle.
    type(ArrayT) :: outArrayH        ! Output array handle.
    type(TableT) :: inTableH         ! Input table handle.
    type(TableT) :: wdxTableH        ! Priority Window Data table dataset handle.
    type(TableT) :: outTableH        ! Output table handle.
    type(TableT) :: modesTableH      ! Modes table dataset handle.
    type(ColumnT), dimension(9) :: inColumnH    ! Input column handle.
    type(ColumnT), dimension(2) :: wdxColumnH   ! Priority Window Data column handle.
    type(ColumnT), dimension(9) :: outColumnH   ! Column handle.
   
   type(ProposalInfoType) :: propInfo   ! Proposal info data.

    character(3) :: fileType = ""    ! Input file file type.
    character(10) :: oswId   = ""       ! Observation Science Window identifier.
    character(20) :: object   = ""    ! Name of observed object.
    character(80) :: observer  = ""   ! Observer's name.
    character(255) :: set     = ""   ! OM ODF Image / Event List / 
!                   Tracking History Data Auxiliary filename.
    character(255) :: pehSet  = ""    ! OM ODF Periodic Housekeeping filename.
    character(255) :: nphSet   = ""   ! OM ODF Non-Periodic Housekeeping filename.
    character(255) :: wdxSet   = ""   ! OM ODF Priority Window Data Auxiliary filename.
    character(255) :: outSet   = ""   ! OM image / event list intermdiate output filename.
    character(255) :: telescope = ""  ! Various keywords.
    character(255) :: instrument = "" 
    character(255) :: dataType = ""    
    character(255) :: obsId = ""     
    character(255) :: expId = ""      
    character(255) :: dateObs = ""    
    character(255) :: dateEnd = ""   
    character(255) :: detSec = "" 
    character(255) :: detsize = ""    
    character(4)   :: exposureIdentifier = ""
    logical :: binBPE = .FALSE.     ! BPE binning flag.

    integer(int16), pointer :: ftFineP(:)   ! Fine time column pointer.
    integer(int16), pointer :: rawXP(:)     ! Raw x column pointer.
    integer(int16), pointer :: rawYP(:)     ! Raw y column pointer.
    integer(int16), pointer :: rawPHA(:)    ! PHA column (Puls Height Analiser)
                                            ! (need for evselect)
    integer(int32), pointer :: frameOut(:)  ! Frame column pointer.
    integer(int32), pointer :: ftCoarseOut(:) ! Coarse time column pointer.
    integer(int16), pointer :: ftFineOut(:)   ! Fine time column pointer.
    integer(int16), pointer :: rawXOut(:)     ! Raw x column pointer.
    integer(int16), pointer :: rawYOut(:)    ! Raw y column pointer.

    integer(int16), pointer :: int16ColumnP(:) ! Column pointer.
    integer(int32) :: mode = 0                
    integer  :: instrumentMode = IMAGING ! Instrument mode (IMAGING, DATA_FAST, etc.)

    integer(int32), pointer :: frameP(:)      ! Frame column pointer.
    integer(int32), pointer :: ftCoarseP(:)   ! Coarse time column pointer.
    integer(int32) :: frame0 = 0             ! Frame column zero
    integer(int32) :: ftCoarse0 = 0          ! Coarse time column zero

    integer(int32), pointer :: nggsP(:)  ! No. good guide stars column pointer.
    integer(int32), pointer :: dxP(:)    ! Drift in X column pointer.
    integer(int32), pointer :: dyP(:)    ! Drift in Y column pointer.
    integer(int32), pointer :: rollP(:)  ! Roll column pointer.
    integer(int32), pointer :: qualityP(:) ! Quality column pointer.
    integer(int32), pointer :: gsXP(:, :)  ! Guide Star X coordinate column pointer.
    integer(int32), pointer :: gsYP(:, :)  ! Guide Star Y coordinate column pointer.
    integer(int32), pointer :: gsCtsP(:, :) ! Guide Star Counts column pointer.
    integer(int32), pointer :: binAx1P(:, :) ! Binax1 pointer.
    integer(int32), pointer :: binAx2P(:, :)  ! Binax2 pointer.
    integer(int32) :: filterAttribute = 0 ! keyword for the filter attribute
    integer(int32) :: bFastId = 1 ! fits attribute for the fast mode
    integer(int32) :: sampTime = 256 ! sampling time to be propagated from input to output
    integer(int32) :: chain  = 0        ! chain flag
    integer(int32) :: tmExpId = 0        ! exposure identifier
    integer(int32) :: sciWin = 0          ! flag for the science window
    integer(int32) :: frmTime = 0        ! frame time 
    integer(int32) :: windowX0 = 0        ! window orinin in X
    integer(int32) :: windowY0 = 0      ! window origin in Y
    ! VNY 29.06.2004
    integer(int32) :: binnedX0 = 0
    integer(int32) :: binnedY0 = 0
    integer(int32) :: binnedDX = 0
    integer(int32) :: binnedDY = 0
    !
    integer(int32) :: windowDX = 0        ! window width
    integer(int32) :: windowDY = 0      ! window height
    integer(int32) :: binX = 1  ! Binning factor for X 
    integer(int32) :: binY = 1    ! Binning factor for Y
    integer(int32), pointer :: int32ImageP(:, :)  ! Integer array pointer.
    integer(int32), pointer :: int32ImageEng4P(:, :, :)  ! Integer array pointer.
    integer(int32), dimension(2) :: imageDim        ! Image dimension.
    !integer(int32), dimension(3) :: imageDimEng4   
    integer(int32) :: nEvents = 0 ! Number of events 
    integer(int32) :: nEventsOut = 0     ! Number of events.
    integer(int32) :: nFrames = 0                 ! Number of tracking frames.
    integer(int32) :: binAx1 = 0                 ! DPU x-axis binning factors.
    integer(int32) :: binAx2 = 0                ! DPU y-axis binning factors.
    integer(int32) :: binax1old = 0
    integer(int32) :: binax2old = 0  
    integer(int32) :: xAxisLength = 0           ! Length of image x-axis.
    integer(int32) :: yAxisLength = 0           ! Length of image y-axis.
    integer(int32) :: position  = 0
    integer(int32), dimension(1) :: columnDim ! Column dimension.
    integer :: imageDimensions(2) 
    integer :: imageDimensionsEng4(3)
    integer   :: expNumber   = 0 ! Exposure Number.
    integer(int32)   :: length1  = 1024 ! String length.
    integer(int32)   :: length2  = 1024  ! String length.
 

    real(single), pointer :: real32ImageP(:, :)  ! Real array pointer.
    real(single) :: exposure = 0.0    ! Exposure length.
    real(double) :: OBTstart = 0.d0   ! Start time.
    real(double) :: OBTend = 0.d0    ! End time.
 

    type(SpacecraftAttitudeType) :: attitude  ! Attitude information

    character(32)   :: filterString= " " ! Exposure filter string.
    !integer(kind=int32)   :: filterIdInt
    character(len=8)  :: omFilter = " "  ! OM filter name                       
    integer(int32) :: blank  = 0 
    real, parameter :: radian = 57.29577951E0 ! Converts radians to degrees
    real(double)    :: attrarad = 0.0d0               ! RA of Attitude in radians
    real(double)    :: attdecrad = 0.0d0              ! DEC of Attitude in radians
    real(double)    :: attPArad = 0.0d0               ! PA of Attitude in radians
    real(double)    :: frameTime = 0.0d0
    real(double)    :: deadFrac = 0.0d0
    integer :: modeFlag = 0   ! 0 = omichain, 1=fast mode, 2 = slew mode, 3 = tracking history mode 
    integer :: noColumns = 0
    ! data or imaging data.
    logical  :: dummyTHXFlag = .FALSE. ! Flag to indicate the absence of
    ! a tracking history file.
   
    
    integer :: i = 0
    integer :: j = 0
    integer :: filtindex = 0
    character(3) :: exposureNumber = "" ! exposure number in string format
    character(9) :: filtdata(0:11) 
    integer :: ixsize = 0
    integer :: iysize = 0
    integer :: switchFT0 = 0
    integer :: switchFT1 = 0
    integer :: switchAux = 0
    integer :: switchCount = 0
    logical :: simulationFlag = .FALSE.

    ! 2011-03-15 vny: variables for testing RAF and AHF settings of SAS_ATTITUDE variable
     real(double)  :: deltaT = 0.0d0
     real(double)  :: currentT = 0.0d0
     real(double)  :: attRAinit = 0.0d0
     real(double)  :: attDECinit = 0.0d0
     integer :: nTsteps = 0
     !character(len=1024),pointer :: pathString
     !character(len=1024), target :: pathString2
     character(len=8) :: attString = ""

     !character(len=3),pointer :: attStringP
     !character(len=3),target :: attStringAux

     character(len=1024) :: attitudeVariable =""
     character(len=1024), pointer :: attitudeVariableOld
     character(len=1024), target :: attitudeVariableOldAux

     !integer(kind=int32) :: attFlag
     integer :: rawAttitudeFlag = 0
     real(double)  :: attRAav20 = 0.0d0
     real(double)  :: attRAav = 0.0d0
     real(double)  :: attDecAv20 = 0.0d0 
     real(double)  :: attDecAv = 0.0d0
     real(double)  :: attPAav20 = 0.0d0
     real(double)  :: attPAav = 0.0d0
     real(double) :: sum = 0.0d0
     real(double)  :: sum20 = 0.0d0
     real(double) :: absdec = 0.0d0
     integer :: rd = 0
     integer :: rm = 0
     integer :: dd = 0
     integer :: dm = 0
     real(single) :: rs = 0.0
     real(single) :: ds = 0.0


     integer(int32) :: derivedBinX=0
     integer(int32) :: derivedBinY=0

     !pathString2=""
     !pathString=>pathString2

     !attStringAux=''
     !attStringP=>attStringAux

     attitudeVariable = ''

     attitudeVariableOldAux= ''
     attitudeVariableOld=>attitudeVariableOldAux
     
     simulationFlag= .false. ! not a simulation mode
     filtdata = (/ "WHITE    ", "GRISM2   ", "UVW1     ", "UVM2     ", &
                 "UVW2     ", "GRISM1   ", "BLOCKED  ", "V        ", &
                 "MAGNIFIER", "U        ", "B        ", "BARREDU  " /)
     
     !****************************************************
     !
     ! Read in parameters.
     !****************************************************
     set = stringParameter("set")
     pehSet = stringParameter("pehset")
     nphSet = stringParameter("nphset")
     wdxSet = stringParameter("wdxset")

     outSet = stringParameter("outset")
     modeFlag = intParameter("modeset")   ! Tells omprep what data mode is 
                                         ! (IMAGING, FAST MODE or SLEW mode, 
                                         ! or IMAGING with the GRISM filter, modeset=4)
     rawAttitudeFlag=intParameter("rawattitude")

     if (modeFlag .eq. 11) then
        modeFlag=1 ! fast but in the simulation mode
        simulationFlag=.true.
        write (errstr, *) " Running in the simulated FAST mode 1"      
        call warning(trim(errstr), "MODE")
     endif

     if (modeFlag .eq. 31) then
        modeFlag=3 ! THX but in the simulation mode
        simulationFlag=.true.
        write (errstr, *) " Running in the simulated THX FAST mode 3"      
        call warning(trim(errstr), "MODE")
     endif


     !****************************************
     ! Set the instrument mode from modeFlag
     !****************************************
     select case(modeFlag)
     case(0)
        instrumentMode = IMAGING
     case(1) 
        instrumentMode = DATA_FAST
     case(4)
        instrumentMode = IMAGING
     case default
        instrumentMode = IMAGING
     end select
     ! NB in the IMAGING mode there is a possisbility of GRISM1 or GRISM2 filter (modeFlag=4)

     if(index(trim(set), "DUMMYTHX.FIT") > 0) then
        dummyTHXFlag = .TRUE.
        exposureIdentifier = GetExposureIdentifier(trim(wdxset)) 
        if (exposureIdentifier .eq. "NNNN") then
           write (errstr, *) " Exposure identifier not found (setting to zero)"      
           call warning(trim(errstr), "EXP_ID")
           exposureIdentifier="S000"
           ! Exposure identifier not found")
        endif
        
     else
        dummyTHXFlag = .FALSE.
        exposureIdentifier = GetExposureIdentifier(trim(set)) 
        if (exposureIdentifier .eq. "NNNN") then
           write (errstr, *) "Exposure identifier not found (setting to zero)"      
           call warning(trim(errstr), "EXP_ID")
           exposureIdentifier="S000"
           !call fatal("L309 Exposure identifier not found")
        endif
        
     end if


     THXFLAG: if(.not. dummyTHXFlag) then
        inDataSetH = dataSet(set, read)

       !**************************************
       !  Start of "OMIMI1" if block.
       !**************************************

       OMIMI1: if (hasBlock(inDataSetH, "OMIMI1")) then
          instrumentMode = IMAGING   
          ! OM Imaging Mode Data Image file.
          fileType = "IMI"
          ! Get a handle on the image array extension.
          inArrayH = array(inDataSetH, "OMIMI1", read)


          ! Read in the image array extension keywords.
          call GetCommonArrayData(inArrayH, telescope, instrument, chain, &
                                           dataType, obsId, expId, tmExpId, &
                                           dateObs, dateEnd, exposure, binBPE)

          call GetArrayData2(inArrayH, sciWin, frmTime, windowX0, &
                           windowY0, windowDX, windowDY, &
                           binAx1, binAx2, binBPE)

          ! VNY 16.10.2003 Getting the filter identifier from the file header
          if (modeFlag .eq. 4) then 
             if (hasAttribute(inDataSetH, "FILTER")) then
                omFilter=stringAttribute(inDataSetH, "FILTER")
             else
                if (hasAttribute(inArrayH, "FILTER")) then
                   omFilter=stringAttribute(inArrayH, "FILTER")
                else
                   omFilter="NONE    "
                endif
             endif
             if (trim(omFilter) .eq. "200") then
                omFilter="GRISM2  "
             endif
             if (trim(omFilter) .eq. "1000") then
                omFilter="GRISM1  "
             endif
          endif

          !**********************************************
          ! Get the actual dimensions of the image stored 
          ! int the file
          !*********************************************
          imageDimensions = dimensions(inArrayH)
          xAxisLength = imageDimensions(1)
          yaxisLength = imageDimensions(2) 
          blank = int32Attribute(inArrayH, "BLANK")

          !**************************************
          ! Get a pointer to the image.
          !**************************************
          int32ImageP => int32Array2Data(inArrayH)

          imageDim = dimensions(inArrayH) ! The dimensions of the image
       
          !*****************************************************************
          ! AMK - 22/09/2000
          ! 
          ! We need to check the image for negative counts. If this occurs
          ! we set the pixel to value BLANK as the system already copes 
          ! with this. A system warning is issued.
          ! This must be taken as a temporary solution.
          !
          !*****************************************************************
          call CheckForNegativeCounts(int32ImageP,  imageDim(1) - 1,  imageDim(2) - 1) 

           
       else ! if OMIMI1
          if (hasBlock(inDataSetH, "OMFAE1")) then   ! "OMFAE1" present.
             ! OM Fast Mode Event List file.
             fileType = "FAE"
             noColumns = 5
             instrumentMode = DATA_FAST
           
             ! Get a handle on the event list binary table extension.
             inTableH = table(inDataSetH, "OMFAE1")

             ! Read in the event list binary table extension keywords.

             call GetCommonTableData(inTableH, telescope, instrument, chain, &
                                           dataType, obsId, expId, tmExpId, &
                                           dateObs, dateEnd, exposure, binBPE)
             call GetData2(inTableH, sciWin, frmTime, windowX0, &
                           windowY0, windowDX, windowDY, binBPE)

             !write(*,*) "L502 windoxX0=", windowX0, " windowY0=", windowY0, &
             !     " windowDX=", windowDX, " windowDY=", windowDY, " binBPE+", binBPE 
  
             filterAttribute = int32Attribute(inTableH, "FILTER")

             bFastId = int32Attribute(inTableH, "BFASTID") ! to be propagated to the output
             sampTime = int32Attribute(inTableH, "SAMPTIME")
             

             ! Get handles and pointers on the event list binary table extension columns.
             inColumnH(1) = column(inTableH, "FRAME", read)
             frameP => int32Data(inColumnH(1))
             inColumnH(2) = column(inTableH, "FTCOARSE", read)
             ftCoarseP =>  int32Data(inColumnH(2))
             inColumnH(3) = column(inTableH, "FTFINE", read)
             ftFineP => int16Data(inColumnH(3))
             inColumnH(4) = column(inTableH, "RAWX", read)
             rawXP => int16Data(inColumnH(4))
             inColumnH(5) = column(inTableH, "RAWY", read)
             rawYP => int16Data(inColumnH(5))

             ! Get the number of rows in the event list binary table extension.
             nEvents = numberOfRows(inTableH)

             !--- VNY 11.07.2001 Checking if some of the values in the data are negative
             !     (corrupted data). If so, these lines will be removed from the output 
             !     event list table in order to protects omfchain from crashing
             !---         
             j = 0 ! counter for negative values
             nEventsOut = 0
             ftCoarse0 = ftCoarseP(0)
             frame0 = frameP(0)
             switchFT0=0
             switchFT1=1
             switchAux=0
             switchCount=0
             !
             do i = 0, nEvents -1
                switchFT1=0
                if (i .gt. 0 .and. i .lt. nEvents-1) then
                   if (ftFineP(i) .lt. 0 .and. ftFineP(i-1) .gt. 0 .and. &
                        frameP(i-1) .ne. frameP(i)) then
                      switchAux=-1
                      switchCount=switchCount+1
                   endif
                   if (ftFineP(i) .lt. 0 .and. ftFineP(i+1) .gt. 0 .and. &
                        frameP(i+1) .ne. frameP(i)) then
                      switchFT1=1
                      switchCount=switchCount+1
                   endif
                   if (ftFineP(i) .lt. 0 .and. switchAux .lt. 0) then
                      ftFineP(i)=-ftFineP(i)
                   endif
                   if (switchFT1 .eq. 1) switchAux=0
                else
                   if (i .eq. 0) then
                      if (ftFineP(i) .lt. 0 .and. ftFineP(i+1) .lt. 0 .and. &
                           frameP(i) .eq. frameP(i+1)) then
                         switchAux=-1
                         switchCount=switchCount+1
                         ftFineP(i)=-ftFineP(i)
                      endif
                   endif
                endif

                if (ftCoarseP(i) < 0 .or. ftFineP(i) < 0 .or. rawXP(i) < 0 &
                    .or. rawYP(i) < 0 .or. frameP(i) < 0 &
                     .or. ftCoarseP(i) .gt. ftCoarse0 + 1000000.0 .or. &
                     frameP(i) .gt. frame0 + 1000000.0) then                         
                   j = j + 1
                else 
                   nEventsOut = nEventsOut + 1
                end if
             end do
             if (switchCount .gt. 0) then
                write (errstr, *) "The sign of FTFINE is oscillating (",switchCount, " times) "      
                call warning(trim(errstr), "FTFINE")
             endif
             ! 11.05.2001: rawPHA is necessary for evselect
             allocate (rawPHA(1 : nEventsOut))
             rawPHA = 1

             !---Added by VNY 11.07.2001 columns with eliminated negative values
             allocate (frameOut(1 : nEventsOut))
             allocate (ftCoarseOut(1 : nEventsOut))
             allocate (ftFineOut(1 : nEventsOut))
             allocate (rawXOut(1 : nEventsOut))
             allocate (rawYOut(1 : nEventsOut))
             j = 0
             do i = 0, nEvents -1
                if ( .not. (ftCoarseP(i) < 0 .or. ftFineP(i) < 0 .or. rawXP(i) < 0 .or. rawYP(i) < 0 &
                       .or. frameP(i) < 0 .or. ftCoarseP(i) .gt. ftCoarse0 + 1000000.0 .or. &
                       frameP(i) .gt. frame0 + 1000000.0)) then
                   j = j + 1                
                   frameOut(j) = frameP(i)
                   ftCoarseOut(j) = ftCoarseP(i)
                   ftFineOut(j) = ftFineP(i)
                   rawXOut(j) = rawXP(i)
                   rawYOut(j) = rawYP(i)
                end if
             end do
          else ! if OMFAE1
             if (hasBlock(inDataSetH, "OMTHX1")) then  ! "OMTHX1" present.


                ! OM Tracking History Data Auxiliary file.
                fileType = "THX"
                noColumns = 9
                ! Get a handle on the event list binary table extension.
                inTableH = table(inDataSetH, "OMTHX1")

                ! Read in the tracking history binary table extension keywords.


                call GetCommonTableData(inTableH, telescope, instrument, chain, &
                                           dataType, obsId, expId, tmExpId, &
                                           dateObs, dateEnd, exposure, binBPE)

                frmTime = int32Attribute(inTableH, "FRMTIME")

                !*********************************************************************************
                ! Get handles and pointers on the tracking history binary table extension columns.
                !*********************************************************************************
                inColumnH(1) = column(inTableH, "FRAME", read)
                frameP => int32Data(inColumnH(1))
                inColumnH(2) = column(inTableH, "NGGS", read)
                nggsP => int32Data(inColumnH(2))
                inColumnH(3) = column(inTableH, "DX", read)
                dXP => int32Data(inColumnH(3))
                inColumnH(4) = column(inTableH, "DY", read)
                dYP => int32Data(inColumnH(4))
                inColumnH(5) = column(inTableH, "ROLL", read)
                rollP => int32Data(inColumnH(5))
                inColumnH(6) = column(inTableH, "QUALITY", read)
                qualityP=> int32Data(inColumnH(6))
                inColumnH(7) = column(inTableH, "GSX", read)
                gsXP => int32VectorData(inColumnH(7))
                inColumnH(8) = column(inTableH, "GSY", read)
                gsYP => int32VectorData(inColumnH(8))
                inColumnH(9) = column(inTableH, "GSCTS", read)
                gsCtsP => int32VectorData(inColumnH(9))

                ! Get the number of frames in the tracking history binary table extension.
                nFrames = numberOfRows(inTableH)

             else ! if OMTHX1
                if (hasBlock(inDataSetH, "ZERODRIFT")) then   ! "ZERODRIFT" present.
                  
                   write (errstr, *) "Missing OM Tracking History Data Auxiliary File. Using ZERODRIFT option"      
                   call warning(trim(errstr), "ZERODRIFT")
                else ! if ZERODRIFT
                   
                   OME4I1: if (hasBlock(inDataSetH, "OME4I1")) then
                      !instrumentMode = IMAGING

                      instrumentMode = ENGINEERING4   
                      ! OM Imaging Mode Data Image file.
                      fileType = "E4I"
                      ! Get a handle on the image array extension.
                      inArrayH = array(inDataSetH, "OME4I1", read)

                      ! Read in the image array extension keywords.
                      call GetCommonArrayData(inArrayH, telescope, instrument, chain, &
                           dataType, obsId, expId, tmExpId, &
                           dateObs, dateEnd, exposure, binBPE)

                      call GetArrayData2(inArrayH, sciWin, frmTime, windowX0, &
                           windowY0, windowDX, windowDY, &
                           binAx1, binAx2, binBPE)
 
                      ! Getting the filter identifier from the file header
                      if (modeFlag .eq. 4) then 
                         if (hasAttribute(inDataSetH, "FILTER")) then
                            omFilter=stringAttribute(inDataSetH, "FILTER")
                         else
                            if (hasAttribute(inArrayH, "FILTER")) then
                               omFilter=stringAttribute(inArrayH, "FILTER")
                            else
                               omFilter="NONE    "
                            endif
                         endif
                         if (trim(omFilter) .eq. "200") then
                            omFilter="GRISM2  "
                         endif
                         if (trim(omFilter) .eq. "1000") then
                            omFilter="GRISM1  "
                         endif
                      endif

                      !**********************************************
                      ! Get the actual dimensions of the image stored 
                      ! int the file
                      !*********************************************
                      imageDimensionsEng4 = dimensions(inArrayH)
                      xAxisLength = imageDimensionsEng4(1)
                      yaxisLength = imageDimensionsEng4(2) 
                      blank = int32Attribute(inArrayH, "BLANK")

                      !**************************************
                      ! Get a pointer to the image.
                      !**************************************
                      int32ImageEng4P => int32Array3Data(inArrayH)

                      imageDim(1) = imageDimensionsEng4(1) ! The dimensions of the image
                      imageDim(2) = imageDimensionsEng4(2)
                      ixsize=imageDim(1)
                      iysize=imageDim(2)
                      !write(*,*) "L709 ixsize=", ixsize, " iysize=", iysize

                      !converting  the 2048x2048x1 image into 2048x2048 
                      int32ImageP => int32ImageEng4P(0:ixsize-1,0:iysize-1,0)

                      !*************************************************************************
                      ! Set the following 7 parameters for Engineering-4 data  to what we know 
                      ! they should be (October 25 2005 - SPR 3532))
                      !*************************************************************************
                      binbpe = .false. 
                      binax1 = 0
                      binax2 = 0
                      windowx0 = 0
                      windowy0 = 0
                      windowdx = 2048
                      windowdy = 2048


                      !*****************************************************************
                      ! AMK - 22/09/2000
                      ! 
                      ! We need to check the image for negative counts. If this occurs
                      ! we set the pixel to value BLANK as the system already copes 
                      ! with this. A system warning is issued.
                      ! This must be taken as a temporary solution.
                      !
                      !*****************************************************************
                      call CheckForNegativeCounts(int32ImageP,  imageDim(1) - 1,  imageDim(2) - 1) 


                      !-----------------------------------------------------------
                   else ! if OME4I1
                      ! Incompatible file.
                      write (errstr, *) "Could not find any compatibile blocks in set ", set
                      call fatal("badFileType")
                   
                      !********************************************************
                      !  End of OMIMI1 if block.
                      !********************************************************
                   endif OME4I1 ! OME4I1



                end if !ZERODRIFT
             end if    !OMTHX1
          end if       ! OMFAE1

       end if OMIMI1
    else ! This is for missing THX files
       write (errstr, *) "Missing OM Tracking History Data Auxiliary File. Using ZERODRIFT option"      
       call warning(trim(errstr),"ZERODRIFT")
    end if THXFLAG

    THXFLAG1: if(.not. dummyTHXFlag) then
       
       !***********************************************************
       ! Start of ZERODRIFT if block - new bit added by JFR on 24.3.2000.
       !***********************************************************
       ZERODRIFT:if (hasBlock(inDataSetH, "ZERODRIFT")) then
          
          outDataSetH = dataSet(outSet, create)
          
          outTableH = addTable(outDataSetH, "ZERODRIFT", 1)
          call copyAllAttributes(attributable(inTableH), attributable(outTableH))
          call AddInt32Column2(outTableH, 0, "DRIFT")
          call release(outTableH)
       else
          
          ! Read in OM ODF Periodic Housekeeping file.
          pehDataSetH = dataSet(pehSet, read)

          !******************************************
          ! Start of OMPEH1 if block.
          !******************************************
          OMPEH1: if (hasBlock(pehDataSetH, "OMPEH1")) then
             !*************************************************************************************
             !
             ! Determine if we are deaing with slew data or imaging.
             ! If we processing slew data then just set the binBPE to TRUE, otherwise process as below
             !*************************************************************************************
             if(modeFlag .eq. 2) binBPE = .true.
             
             SLEWFLAG1: if(modeFlag .ne. 2) then


                length1 = len_trim(expId)
                length2 = len_trim(obsId)

                !*****************************************
                ! Set the string for the exposure number
                !****************************************
                exposureNumber = expId(length2 + 1 : length1)

                !****************************************
                ! Now convert the string to an integer
                !*****************************************
                
                read(exposureNumber,'(i3)') expNumber


                !*************************************************************
                ! Now find the accurate start and end times for this exposure.
                !*************************************************************
                if (modeFlag .ne. 3) then

                   if (.not. simulationFlag) then
                      call getstartendexp(nphSet, expNumber, instrumentMode, OBTstart, &
                           OBTend, dateObs, dateEnd, filterString)
                   endif

                    ! VNY 16.10.2003 set the filterString to the value picked up from the input file header
                    if (modeFlag .eq. 4) then
                       if (trim(omFilter) .ne. "NONE") then
                          filterString=omFilter
                       endif
                    endif
 

                    !************************************************************
                    ! Check to see if exposure is zero- if so use OBTend-OBTstart
                    ! as actual exposure time
                    !************************************************************

                    if (exposure .lt. 1e-6) then
                        exposure = OBTend - OBTstart
                        write (errstr, *) "Exposure time is zero- calculating it from the summary file" 
                        call warning(trim(errstr),"Exposure time")
                        if (exposure .lt. 1e-6) then
                           write (errstr, *) "Exposure time from summary file is zero!"
                            call fatal(trim(errstr),"Exposure time")
                        end if

                    end if  
                else
  
!                    position = index(set, "OM")
!                    exposureIdentifier = set(position + 2 : position + 5)

              
                   call OAL_setState(OM, instrumentMode, instrExposureId = exposureIdentifier)
              
                    filterString  = OAL_getIPPVstring("FILTER") 
      
                end if


                if (trim(filterString) .eq. "UNKNOWN" .and. instrumentMode .eq. &
                                    DATA_FAST .and. fileType .eq. "FAE") then

                   filtindex = filterAttribute / 100
                   if (filtindex .eq. 21) filtindex = 22
                   
                   filtindex = filtindex / 2
                   filterString = filtdata(filtindex)
                end if !trim filterString


              
                !*****************************************
                ! Convert the string to upper-case characters
                !*****************************************
                call ConvertToUpperCase(filterString)

                !************************************************
                ! Check that the filter string is okay
                !************************************************


                !-- - call fatal only in the case of the Imaging Mode                 
                if ((IsBadFilter(filterString)) .and. (instrumentMode .ne. DATA_FAST)) then
                    call fatal("Invalid filter string", filterString)
                end if

!                if (fileType .eq. "IMI" .or. fileType .eq. "E4I") &
                if (fileType .eq. "IMI") &
                     call CheckImageSizeOkay(binBPE, windowDX, windowDY, binAx1, &
                     binAx2, xAxisLength, yAxisLength) 

             end if SLEWFLAG1

          else ! Incompatible file.
             write (errstr, *) "Could not find any compatibile blocks in set ", pehSet
             call fatal("badFileType")
             !************************************************
             ! End of OMPEH1 block.
             !***********************************************
          end if OMPEH1

          ! Release handles on data.
          call release(pehDataSetH)

          ! Read in OM ODF Priority Window Data Auxiliary file.
          ! VNY 23.10.2003 setting binax1 and binax2 to zero is wrong

          binax1old=binax1
          binax2old=binax2

          binax1 = 0
          binax2 = 0


          if (fileType .eq. "IMI" .or. fileType .eq. "FAE" .or. fileType .eq. "E4I") then
             
             wdxDataSetH = dataSet(wdxSet, read)
          
             if (hasBlock(wdxDataSetH, "OMWDX1")) then
                ! Get a handle on the window data binary table extension.
                wdxTableH = table(wdxDataSetH, "OMWDX1")

                !****************************************************
                ! binAx1 is found in SWP1 and binAx2 in SWP2.
                ! PSW sciWin numbering system still not settled.
                ! May add sampling time and bfast ID to event list.
                ! The meaning of SWP1 and SWP2 is mode-dependent;
                ! in the Fast mode this is the sampling time in 
                ! units of 1/1024 s
                !****************************************************

                wdxColumnH(1) = column(wdxTableH, "SWP1", read)
                binAx1P => int32VectorData(wdxColumnH(1))
                binAx1 = binAx1P(sciWin, 0)
                wdxColumnH(2) = column(wdxTableH, "SWP2", read)
                binAx2P => int32VectorData(wdxColumnH(2))
                binAx2 = binAx2P(sciWin, 0)

                !write(*,*) "L928 binAx1=", binAx1, " binAx1old=",binax1old

                !  Protection against wrong data in wdx-table
                if (binax1 .gt. 2) then
                   if (fileType .eq. "IMI") then
                   !if (fileType .eq. "IMI" .or. fileType .eq. "FAE") then
                      binax1=binax1old
                      write (errstr, *) "The X-binning factor in the WDX-table seems to be wrong"      
                      call warning(trim(errstr), "correcting BinX")
                   else
                      binax1=0 ! (engineering-4 and Fast (FAE) is always unbinned)
                   endif
                endif
                if (binax2 .gt. 2) then
                   !if (fileType .eq. "IMI" .or. fileType .eq. "FAE") then
                   if (fileType .eq. "IMI") then
                      binax2=binax2old
                      write (errstr, *) "The Y-binning factor in WDX-table seems to be wrong"      
                      call warning(trim(errstr), "correcting BinY")
                   else
                      binax2=0 ! (engineering-4 and Fast is always unbinned)
                   endif

                endif
                if (binax1 .ne. binax2) then
                   binax1=min(binax1,binax2)
                   binax2=min(binax1,binax2)
                endif
                ! Release handles on data.
                call release(wdxColumnH(1))
                call release(wdxColumnH(2))
                call release(wdxTableH)
                call release(wdxDataSetH)
             else ! Incompatible file.
                call fatal("badFileType")
             end if
          end if

          ! Get the proposal info. from observation summary file.
          if (simulationFlag) then
             observer="O.M. Simulation"
             object="Simulated"
          else
             call OAL_proposalInfo(propInfo)
             observer = propInfo%GoName
             object = propInfo%targetName
          endif

          ! Write out OM image / event list / tracking history intermdiate output filename.
          outDataSetH = dataSet(outSet, create)

          call setAttribute(outDataSetH, "TELESCOP", trim(telescope), &
                           "Telescope (mission) name")
          call setAttribute(outDataSetH, "INSTRUME", trim(instrument), &
                           "Instrument name")
          ! Chain tells us which detector is in use.
          if (chain .eq. 1) then
             call setAttribute(outDataSetH, "DETNAM", "REDUNDANT", "PRIME or REDUNDANT")
          else
             call setAttribute(outDataSetH, "DETNAM", "PRIME", "PRIME or REDUNDANT")
          end if

          call setAttribute(outDataSetH, "DATATYPE", trim(dataType), "Instrument mode")
         
          if (trim(dataType) .eq. "IMG.IM" .or. trim(dataType) .eq. "ENG.IM") then
             call setAttribute(outDataSetH, "DATAMODE", "IMAGING", "Instrument mode")
          else
             if (trim(dataType) .eq. "FST.EL") then
                call setAttribute(outDataSetH, "DATAMODE", "FAST", "Instrument mode")
             else
                if (trim(dataType) .eq. "TRH.EL") then
                   call setAttribute(outDataSetH, "DATAMODE", "TRACKING", "Instrument mode")
                else
                   call fatal("wrongMode", "OM ODF file not taken in imaging/fast/tracking mode")
           
                endif
             end if
          end if
          ! Write FILTER keyword to outSet.
          call setAttribute(outDataSetH, "FILTER", trim(filterString), "Filter ID")

          ! Write keywords to outSet.

          call setAttribute(outDataSetH, "OBS_MODE", "POINTING", "Observation mode (pointing or slew)")
          call setAttribute(outDataSetH, "OBS_ID", trim(obsId), "Observation identifier")
          call setAttribute(outDataSetH, "EXP_ID", trim(expId), "Exposure identifier")
      
          call setAttribute(outDataSetH, "ORIGIN", "UNKNOWN", "Origin of FITS file")
          call setAttribute(outDataSetH, "OBJECT", trim(object), "Name of observed object")


          if(modeFlag .ne. 2) then 
             ! Get attitude information

          if (.not. simulationFlag) then

             !if ((rawAttitudeFlag .eq. 0) .and. (trim(attitudeVariableOld) .eq. "RAF" &
             !     .or. trim(attitudeVariableOld) .eq. "raf") .or. &
             !     rawAttitudeFlag .gt. 0) then 
             !   ! using RAF
             !   call OAL_setState(instrument = OM, dataMode = instrumentMode, &
             !        instrExposureId =  exposureIdentifier, &
             !        attitudeFromAhf = .false.)
             !else
             !   ! using AHF
                call OAL_setState(instrument = OM, dataMode = instrumentMode, &
                     instrExposureId =  exposureIdentifier, &
                     attitudeFromAhf = .true.)
                ! vny 2011-03-18: if attitudeFromAhf is set to false, then OAL gives an error message
             !endif

             call GetStartTime

             if (modeFlag .ne. 3) then
                ! 2011-03-15 Check the stability of the attitude
                !write(*,*) "L973 dateObs=", trim(dateObs)
                !write(*,*) "L974 dateEnd=", trim(dateEnd)
                !write(*,*) "L975 toTimeTag(dateObs)=",toTimeTag(dateObs) 
                !write(*,*) "L976 toTimeTag(dateObs)=",toTimeTag(dateEnd) 
                !write(*,*) "L977 OBTstart=", OBTstart, " OBTend=", OBTend
                !write(*,*) "L978 exposure=", exposure, " diff=", OBTend-OBTstart
             
                !write(*,*) "L997 rawAttitudeFlag=", rawAttitudeFlag

                deltaT=0.5_double
                nTsteps=0
                if (exposure .gt. 10.0) then
                   nTsteps=int(exposure/deltaT+0.5)
                endif

                !call getsasattitude(pathString)
                !write(*,*) "L993 pathString=", trim(pathString)

 
                !attitudeVariableOld=''
                call getsasattitude(attitudeVariableOld)
                !write(*,*) "L1019 attitudeVariableOld=", trim(attitudeVariableOld)

                if ((rawAttitudeFlag .eq. 0) .and. (trim(attitudeVariableOld) .eq. "RAF" &
                     .or. trim(attitudeVariableOld) .eq. "raf") .or. &
                     rawAttitudeFlag .gt. 0) then 

                   if (rawAttitudeFlag .eq. 1) then
                      !write (errstr, *) "Using data from the Raw Attitude File (RAF)" 
                      call message(AppMsg, Verbose,"Using data from the Raw Attitude File (RAF)")
                      !call warning(trim(errstr), "averaged over the first 20 s of the exposure")
                      call message(AppMsg, Verbose,"averaged over the first 20 s of the exposure")
                   else
                      !write (errstr, *) "Using data from the Raw Attitude File (RAF)"      
                      !call warning(trim(errstr), "averaged over the whole exposure time")
                      call message(AppMsg, Verbose, "Using data from the Raw Attitude File (RAF)")
                      call message(AppMsg, Verbose, "averaged over the whole exposure time")
                   endif

                   ! set the system variable SAS_ATTITUDE to RAF
                   attString="RAF"//char(0)
                   call setsasattitude(attString)
                   !call setsasattitudeahf
          

                   if(OAL_getAttitude(OBTstart, attitude)) then
                      attRAinit = attitude%ra       ! The ra of the spacecraft
                      attDECinit = attitude%dec     ! The dec of the spacecraft
                      attPArad = attitude%apos     ! The roll-angle of the spacecraft  
                      !write(*,*) "L1025  raInit=", attRAinit, " decInit=", attDECinit
                   end if

                   attRAav20=0.0_double
                   attRAav=0.0_double
                   attDECav20=0.0_double
                   attDECav=0.0_double
                   attPAav=0.0_double
                   attPAav20=0.0_double
                   sum=0.0_double
                   sum20=0.0_double

                   do i=1,nTsteps
                      currentT=OBTstart+dble((i-1)*deltaT)
                   
                      if(OAL_getAttitude(currentT, attitude)) then
                         attrarad = attitude%ra       ! The ra of the spacecraft
                         attdecrad = attitude%dec     ! The dec of the spacecraft
                         attPArad = attitude%apos     ! The roll-angle of the spacecraft  
                         !write(*,*) "L998 i=", i, " t=", currentT-OBTstart, " dra=", &
                         !     (attrarad-attRAinit)*206265.0_double, " ddec=", (attdecrad-attDECinit)*206265.0_double

                         attRAav=attRAav+attrarad
                         attDecAv=attDecAv+attdecrad
                         attPAav=attPAav+attPArad
                         sum=sum+1.0_double
                         if (i .le. 40) then
                            attRAav20=attRAav20+attrarad
                            attDecAv20=attDecAv20+attdecrad
                            attPAav20=attPAav20+attPArad
                            sum20=sum20+1.0_double
                            !write(*,'(f6.1,f8.2,f8.2)') currentT-OBTstart,  &
                            !     (attrarad-attRAinit)*206264.806_double, (attdecrad-attDECinit)*206264.806_double
                         endif
                      end if
                   enddo
 
                   if (sum20 .gt. 1.0_double) then
                      attRAav20=attRAav20/sum20
                      attDecAv20=attDecAv20/sum20
                      attPAav20=attPAav20/sum20
                   endif
                   if (sum .gt. 1.0_double) then
                      attRAav=attRAav/sum
                      attDecAv=attDecAv/sum
                      attPAav=attPAav/sum
                   endif

                   ! restore the old value of SAS_ATTITUDE
                   !attString=attitudeVariableOld//char(0)
                   !call setsasattitude(attString)

                   if (rawAttitudeFlag .le. 1)then
                      attrarad=attRAav20
                      attdecrad=attDecAv20
                      attPArad =attPAav20
                   else
                      attrarad=attRAav
                      attdecrad=attDecAv
                      attPArad =attPAav
                   endif
                else
                   attrarad = 0.0       ! The ra of the spacecraft
                   attdecrad = 0.0     ! The dec of the spacecraft
                   attPArad = 0.0     ! The roll-angle of the spacecraft
               
                   !write (errstr, *) "Using data from the Attitude History File (AHF)"      
                   !call warning(trim(errstr), "the first value")   
                   call message(AppMsg, Verbose, "Using  the first point of the Attitude History File (AHF)")   
                   if(OAL_getAttitude(toTimeTag(dateObs), attitude)) then
                      attrarad = attitude%ra       ! The ra of the spacecraft
                      attdecrad = attitude%dec     ! The dec of the spacecraft
                      attPArad = attitude%apos     ! The roll-angle of the spacecraft  
                      ! call OAL_toAttitudeMatrix(attitude,matrix)
                      !write(*,*) "L1132 dateObs attrarad=", attrarad*206264.806_double, &
                      !     " attdecrad=", attdecrad*206264.806_double
                   end if

                endif
             else
                attrarad = 0.0       ! The ra of the spacecraft
                attdecrad = 0.0     ! The dec of the spacecraft
                attPArad = 0.0     ! The roll-angle of the spacecraft
                 
                if(OAL_getAttitude(toTimeTag(dateObs), attitude)) then
                   attrarad = attitude%ra       ! The ra of the spacecraft
                   attdecrad = attitude%dec     ! The dec of the spacecraft
                   attPArad = attitude%apos     ! The roll-angle of the spacecraft  
                end if

             endif

          else

             attrarad = 0.0       ! The ra of the spacecraft
             attdecrad = 0.0     ! The dec of the spacecraft
             attPArad = 0.0     ! The roll-angle of the spacecraft                 
          endif
       end if
       !write(*,'(f6.1,f8.2,f8.2)') attPArad*206264.806_double,  &
       !     attrarad*206264.806_double, attdecrad*206264.806_double
       !write(*,*) "PosAng=",attPArad/206264.806_double,  &
       !     "RA=",attrarad/206264.806_double, "DEC=",attdecrad/206264.806_double


       rd=int(attrarad*radian)
       rm=int((attrarad*radian-real(rd))*60.0)
       rs=((attrarad*radian-real(rd))*60.0-real(rm))*60.0

       absdec=abs(attdecrad*radian)
       dd=int(absdec)
       dm=int((absdec-real(dd))*60.0)
       ds=((absdec-real(dd))*60.0-real(dm))*60.0
       if (attrarad .lt. 0.0) then
          dd=-dd
       endif

       write(errstr,'(a,i4.3,i3.2,f5.1,a,i3.2,i3.2,f6.2,a,f7.2)')  "RA =",rd,rm,rs, &
            "     DEC =",dd,dm,ds, "     PA =",attPArad*radian
       !call warning("attitude data", trim(errstr))  
       call message(AppMsg,Verbose)  
          !**********************************************************************
          ! Positions refer to spacecraft pointing and not the instrument.
          !**********************************************************************
          call setAttribute(outDataSetH, "RA_SCX", attrarad*radian, &
                        "degrees", "RA of spacecraft pointing")
          call setAttribute(outDataSetH, "DEC_SCX",attdecrad*radian, &
                        "degrees", "DEC of spacecraft pointing")
          call setAttribute(outDataSetH, "PA", attPArad*radian, &
                        "degrees", "ROLL angle of spacecraft pointing")


          !****************************************
          ! Add rest of keywords to outSet.
          !****************************************
          call setAttribute(outDataSetH, "RADECSYS", "FK5", " ")
          call setAttribute(outDataSetH, "EQUINOX", 2000, "", "Equinox of celestial coordinate system")
          call setAttribute(outDataSetH, "EXPOSURE", real(exposure), "secs", "Exposure")
          call setAttribute(outDataSetH, "DATE-OBS", trim(dateObs), "Date observations were made")
          call setAttribute(outDataSetH, "DATE-END", trim(dateEnd), "Date observations ended")
          call setAttribute(outDataSetH, "OBSERVER", observer, "Name of PI")

          if (fileType .eq. "IMI" .or. fileType .eq. "FAE" &
               .or. fileType .eq. "E4I") call  AddTimingAttributes(outDataSetH)
        
          ! initialise binning    
! VNY 23.10.2003 returning binax1 and binax2 to the old algoritm, but it is wrong
          binx = 2 ** binax1
          biny = 2 ** binax2

!          binx=binax1
!          biny=binax2
!          if (binax1 .eq. 0) then 
!             binx=1.0
!             binax1=1
!          endif
!          if (binax2 .eq. 0) then
!             biny=1.0
!             binax2=1
!          endif

          if (trim(dataType) .eq. "FST.EL") then
             binx = 1
             biny = 1
          end if

          if (binBPE) then
! VNY 23.10.2003 returning binax1 and binax2 to the old algorithm, but it is wrong
             binx = binx * 2.0
             biny = biny * 2.0
!             binx = 2.0
!             biny= 2.0
          end if


          !write(*,*) "L1271 binx=", binx, " biny=", biny
          !write(*,*) "L1272 windoxX0=", windowX0, " windowY0=", windowY0, &
          !     " windowDX=", windowDX, " windowDY=", windowDY, " binBPE+", binBPE 
          !write(*,*) "L1280 ixsize=", ixsize, " iysize=", iysize
          !write(*,*) "L1281 binAx1=", binAx1, " binAx1old=",binax1old

          derivedBinX=0
          if (ixsize .gt. 0) then 
             derivedBinX=windowDX/ixsize
          endif
          derivedBinY=0
          if (iysize .gt. 0) then
             derivedBinY=windowDY/iysize
          endif

          !write(*,*) "L1297 derivedBinX=", derivedBinX, " derivedBinY=", derivedBinY
          ! pritect the binning factor 
          if (derivedBinX>0) then
             if (binx .ne. derivedBinX) then
                !write (errstr, *) "The binning factor ",binx, " seems to be wrong (fixing) "      
                !call warning(trim(errstr), "binning")
                write (errstr, *) "The binning factor ",binx, " seems to be wrong (fixing) "      
                call warning("binning",trim(errstr))
                binx=derivedBinX
                biny=derivedBinX
                binAx1=0
                binAx2=0
             endif             
          endif

          binnedX0=(windowX0+mod(windowX0,binx))/binx
          binnedY0=(windowY0+mod(windowY0,biny))/biny
          binnedDX=windowDX/binx
          binnedDY=windowDY/biny


          ! dataType for Engineering-4 is ENG.IM (not E4I.IM
          !--------------------------------------------------------------------------------
          if (trim(dataType) .eq. "IMG.IM" .or. trim(dataType) .eq. "FST.EL" &
               .or. trim(dataType) .eq. "ENG.IM") then
             write(oswId, '(i10)') sciwin
             write (detSec, '(a1,i4,a1,i4,a1,i4,a1,i4,a1)') "[",binnedX0, ":", binnedX0 + binnedDX - 1, &
                  & ",",binnedY0, ":", binnedY0 + binnedDY - 1, "]" 
             write (detsize,'(a3,i4,a3,i4,a1)') "[1:", INT(2048.0/real(binx)),",1:",INT(2048.0/real(biny)), "]"
             call setAttribute(outDataSetH, "DETSIZE", trim(detsize), "Mosaic Keywords")
             call setAttribute(outDataSetH, "DETSEC", trim(detSec), "Mosaic Keywords")
             call setAttribute(outDataSetH, "OSW_ID", trim(oswId), "Identifier for current OSW")
          end if
          call setAttribute(outDataSetH, "BINBPE", binBPE, "BPE binning enabled")
 
          if (trim(dataType) .eq. "IMG.IM" .or. trim(dataType) .eq. "ENG.IM") then

             call setAttribute(outDataSetH, "MISSPIX", real(blank), "Value indicating missing pixel")
             call setAttribute(outDataSetH, "BINAX1", binAx1, "", "CCD x-axis binning 2**N")
             call setAttribute(outDataSetH, "BINAX2", binAx2, "", "CCD y-axis binning 2**M")
          end if
          if(trim(dataType) .eq. "FST.EL") then
             call setAttribute(outDataSetH, "BINAX1", 0, "", "CCD x-axis binning 2**N")
             call setAttribute(outDataSetH, "BINAX2", 0, "", "CCD y-axis binning 2**M")      
             
             call setAttribute(outDataSetH, "BFASTID", bFastId, "", "Blue dsp 1 or 2") 
             
             call setAttribute(outDataSetH, "SAMPTIME", &
                  sampTime, "Sample time (1/1024 secs)", "Fast Mode slice time") 

          end if
          call setAttribute(outDataSetH, "FRMTIME", frmTime, "1/1024 secs", "Tracking frame time")
          call setAttribute(outDataSetH, "TMEXP_ID", tmExpId, "", "Exposure identifier")
          call setAttribute(OutDataSetH, "HDUCLASS", "OGIP", "Format conforms to OGIP/GSFC conventions")

          !***********************************************************
          ! Start of IMG.IM if block.
          !*********************************************************** 
          IMGIM: if (trim(dataType) .eq. "IMG.IM" .or. trim(dataType) .eq. "ENG.IM") then
             ! Convert from int32 data in odf to real32 data for output.
             outArrayH = addArray(outDataSetH, "PRIMARY", real32, imageDim)
             real32ImageP => real32Array2Data(outArrayH)
             real32ImageP = real(int32ImageP)
             call setAttribute(outArrayH, "WINDOWX0", windowx0)
             call setAttribute(outArrayH, "WINDOWDX", windowdx)
             call setAttribute(outArrayH, "WINDOWY0", windowy0)
             call setAttribute(outArrayH, "WINDOWDY", windowdy)



! VNY 16.10.2003
             if (modeFlag .ne. 4) then
                outArrayH = addArray(outDataSetH, "RAW", real32, imageDim)
                real32ImageP => real32Array2Data(outArrayH)
                real32ImageP = real(int32ImageP)
             endif
          else
             if (trim(dataType) .eq. "FST.EL") then
                outTableH = addTable(outDataSetH, "OMFAE1", nEventsOut)
                call copyAllAttributes(attributable(inTableH), attributable(outTableH))

                !**********************************************************************
                !---VNY 11.07.2001 Output table is being reduced if negative values were
                !  found in the input data
                !**********************************************************************
                call AddInt32Column1(outTableH, frameOut, "FRAME")
                call AddInt32Column1(outTableH, ftCoarseOut, "FTCOARSE")
                call AddInt16Column1(outTableH, ftFineOut, "FTFINE")
              
                columnDim(1) = 1
                call  AddInt16Column2(outTableH, rawPHA, "PHA", "", columnDim) 


                outColumnH(4) = addColumn(outTableH, "RAWX", integer16)
                int16ColumnP => int16Data(outColumnH(4))
                int16ColumnP = rawXOut

                call setAttribute(outColumnH(4), "TLMIN", windowX0 / binx)
                call setAttribute(outColumnH(4), "TLMAX", (windowX0 + windowDX) / binx - 1)

                outColumnH(5) = addColumn(outTableH, "RAWY", integer16)
                int16ColumnP => int16Data(outColumnH(5))
                int16ColumnP = rawYOut

                !-----TLMIN, TLMAX have been changed by V.Ye. 11.05.2001 in order evselect to be able to construct the image
 

                !------V.Ye. 11.05.2001 column 6 (PHA - pulse Height Analiser) has been added to the output table
                !                       in order evselect to be able to construct the time series 
             
             

                deallocate (rawPHA)
                deallocate (frameOut)
                deallocate (ftCoarseOut)
                deallocate (ftFineOut)
                deallocate (rawXOut)
                deallocate (rawYOut)

                call setAttribute(outColumnH(5), "TLMIN", windowY0/biny)
                call setAttribute(outColumnH(5), "TLMAX", (windowY0+windowDY)/biny-1)
                
                call release(outColumnH(4))
                call release(outColumnH(5))


             else
                
                if (trim(dataType) .eq. "TRH.EL") then

                   outTableH = addTable(outDataSetH, "OMTHX1", nFrames)

                   call copyAllAttributes(attributable(inTableH), attributable(outTableH))

                 
                   !******************************************************
                   ! Reset the start and end times to the correct values
                   !******************************************************
                   call setAttribute(outTableH, "DATE-OBS", trim(dateObs), "Start time of exposure")
                   call setAttribute(outTableH, "DATE-END", trim(dateEnd), "End time of exposure")

                   columnDim(1) = 1
                   call AddInt32ArrayColumn(outTableH, frameP, "FRAME", "COUNTER", columnDim) 
                   call AddInt32ArrayColumn(outTableH, nggsP, "NGGS", "COUNTS", columnDim) 
                   call AddInt32ArrayColumn(outTableH, dXP, "DX", " ", columnDim) 
                   call AddInt32ArrayColumn(outTableH, dYP, "DY", " ", columnDim) 
                   call AddInt32ArrayColumn(outTableH, rollP, "ROLL", " ", columnDim) 
                   call AddInt32ArrayColumn(outTableH, qualityP, "QUALITY", " ", columnDim) 

                   columnDim(1) = size(gsXP)/size(nggsP)
                   call AddInt32Array2Column(outTableH, gsXP, "GSX", "PIXEL", columnDim) 
                   call AddInt32Array2Column(outTableH, gsYP, "GSY", "PIXEL", columnDim) 
                   call AddInt32Array2Column(outTableH, gsCtsP, "GSCTS", "COUNTS", columnDim) 
                   
                   !********************************************
                   ! End of IMG.IM if block.
                   !*******************************************
                end if
             end if
          end if IMGIM

          !************************************
          ! Start of IMG_FST if block
          !************************************
          IMG_FST: if (trim(dataType) .eq. "IMG.IM" .or. trim(dataType) .eq. "FST.EL" &
               .or. trim(dataType) .eq. "ENG.IM") then

             !---VNY 05.07.2001 the next line should be checked
             instrumentMode = DATA_FAST
             ! VNY 17.02.2005 
             if (trim(dataType) .eq. "IMG.IM") then
                instrumentMode= IMAGING
             endif

             if (trim(dataType) .eq. "ENG.IM") then
                instrumentMode= ENGINEERING4
             endif

             ! Add a MODES extension, detailing the science window parameters.
             modesTableH = addTable(outDataSetH, "MODES", 1)

             columnDim(1) = 1

             if (trim(dataType) .eq. "IMG.IM" .or. trim(dataType) .eq. "ENG.IM") then
                mode = 0
             else
                if (trim(dataType) .eq. "FST.EL") then
                   mode = 1
                endif
             end if

             call AddInt32Column(modesTableH, mode, "MODE", "none", columnDim) 
             call AddInt32Column(modesTableH, windowX0, "WINDOWX0", "pixels", columnDim) 
             call AddInt32Column(modesTableH, windowY0, "WINDOWY0", "pixels", columnDim) 
             call AddInt32Column(modesTableH, windowDX, "WINDOWDX", "pixels", columnDim) 
             call AddInt32Column(modesTableH, windowDY, "WINDOWDY", "pixels", columnDim) 
             call release(modesTableH)
             !*****************************************************
             ! End of IGF_FST if block
             !*****************************************************
          end if IMG_FST

          !*********************************************************
          ! End of ZERODRIFT if block.
          !*********************************************************
       end if  ZERODRIFT
    else
       outDataSetH = dataSet(outSet, create)
       outTableH = addTable(outDataSetH, "ZERODRIFT", 1)

       call AddInt32Column2(outTableH, 0, "DRIFT")
       call release(outTableH)
    end if THXFLAG1
    if (fileType .eq. "IMI" .or. fileType .eq. "E4I") then
       call release(outArrayH)
       call release(inArrayH)
    end if
    if (fileType .eq. "FAE" .or. fileType .eq. "THX") then
        do i = 1,  noColumns 
            call release(inColumnH(i))
        end do
        call release(inTableH)
        call release(outTableH)
    end if
 
    if(.not. dummyTHXFlag) call release(inDataSetH)
    !************************************************
    ! 20.2.2002 - Add the new attributes for SCR 84
    !************************************************
    if (.not. simulationFlag) then
       call OAL_addCommonAttributes(outDataSetH)
    endif
    !************************************************
    ! 20.2.2002 - Add the new attribute "EXPIDSTR"-
    ! the four-character exposure-identifier
    !************************************************
    
    ! RDS - SPR-7078 - report position of LAST occurrence of "_OM"
    position = index(set, "_OM", .true.)

    !exposureIdentifier = set(position + 2 : position + 5)
    exposureIdentifier = set(position + 3 : position + 6)

    call setAttribute(outDataSetH, "EXPIDSTR", exposureIdentifier, "Exposure identifier")
    if (modeFlag .ne. 3) then

        !***************************************************************
        ! 15th August 2001 - Retrieve the CCD frame-time (milli-secs)  & 
        ! dead fraction and store in the image header for use by other programs
        !***************************************************************
        wdxDataSetH = dataSet(wdxSet, read)
          

        call getFrameTime(wdxDataSetH, frameTime, deadFrac, binBPE)  ! Frame time is in milli-seconds
        call release(wdxDataSetH)
   
        !************************************************
        ! Add CCD frame-time and dead-fraction to header
        !************************************************
        call setAttribute(outDataSetH, "FRAMTIME", frameTime, "millisecs", "CCD frame time")

        call setAttribute(outDataSetH, "DEADFRAC", deadfrac, "", "CCD dead fraction")
    end if
    call release(outDataSetH)

    !**************************************************************
    ! AMK - 21/07/2000
    !
    ! Added. Need to create a dummy tracking file if in slew mode
    !**************************************************************
    if(modeFlag .eq. 2) call CreateDummyTrackingHistoryFile
  
end subroutine omprep
!*********************************************
! ConvertToUpperCase - Converts any lower-case
! characters in a string to upper-case characters.
!*********************************************
subroutine ConvertToUpperCase(string)
    character(*), intent(inout) :: string
    character(1) :: cha
    integer i, newchar

    do i = 1, len(string)
        cha = string(i : i)
        if (lge(cha, 'a') .and. lle(cha, 'z')) then
            newchar = ichar(cha) - ichar('a') + ichar('A')
    
            string(i : i) = char(newchar)
        end if
    end do
end subroutine ConvertToUpperCase
!************************************************
! IsBadFilter - Checks to see if  a given filter is
! an invalid one- badFilter set to  TRUE if so- FALSE
! otherwise
!************************************************
logical function IsBadFilter(filterString)
    character(len=*), intent(inout) :: filterString
    
    select case(trim(filterString))

!        case ("BLOCKED")
!            IsBadFilter = .true.

        case("UNKNOWN")
            IsBadFilter = .true.

!        case("GRISM2")
!            IsBadFilter = .true.

!        case("GRISM1")
!            IsBadFilter = .true.

        case default
            IsBadFilter = .false.

    end select

end function IsBadFilter
!***************************************************************
! subroutine AddTimingAttributes
!**************************************************************
subroutine AddTimingAttributes(dataSetH)
   type(DataSetT), intent(in) :: dataSetH

   call setAttribute(dataSetH, "TIMESYS", "TT", "XMM time will be TT (Terrestial time)")
   call setAttribute(dataSetH, "MJDREF", 50814.0, "d", "1998-01-01T00:00:00 (TT) expressed in MJD (TT)")
   call setAttribute(dataSetH, "TIMEZERO", 0, "s", "Clock correction (if not zero)")
   call setAttribute(dataSetH, "TIMEUNIT", "s", "All times in s unless specified otherwise")
   call setAttribute(dataSetH, "CLOCKAPP", "T", "Clock correction applied")
   call setAttribute(dataSetH, "TIMEREF", "LOCAL", "Reference location of photon arrival times")
   call setAttribute(dataSetH, "TASSIGN", "SATELLITE", "Location of time assignment")

end subroutine AddTimingAttributes
!*********************************************
! Subroutine  CheckImageSizeOkay
! Checks to see if the image size corresponds
! to what the program expects it to be. 
!*********************************************
subroutine CheckImageSizeOkay(binBPE, windowDx, windowDy, &
                  binXaxis, binYaxis, xAxisLength, yAxisLength)
    logical, intent(in) :: binBPE
    integer(int32), intent(in) :: windowDx
    integer(int32), intent(in) :: windowDy
    integer(int32), intent(in) :: binXaxis
    integer(int32), intent(in) :: binYaxis
    integer(int32), intent(in) :: xAxisLength
    integer(int32), intent(in) :: yAxisLength
    integer(int32) :: factor

    if (binBPE) then
        factor = 2
    else
        factor = 1
    end if

    if ((windowDx / (binXaxis * factor)) .ne. xAxisLength) then
        write(errstr, *) "X Image size (", (windowDx / (binXaxis * factor)), &
                         ") computed from BINPE (", binBpe, ") WINDOWDX (", windowDx, &
                         ") and BINXAXIS (", binXaxis, ") does not equal NAXIS1 (", xAxisLength, ")"
        call fatal(errstr)
    end if

    if ((windowDy / (binYaxis * factor)) .ne. yAxisLength) then
       
        write(errstr, *) "Y Image size (", (windowDy / (binYaxis * factor)), &
                         ") computed from BINPE (", binBpe, ") WINDOWDY (", windowDy, &
                         ") and BINYAXIS (", binYaxis, ") does not equal NAXIS2 (", yAxisLength, ")"
  

    end if

end subroutine CheckImageSizeOkay

!********************************************
! subroutine GetData2
!********************************************
subroutine GetData2(tableH, sciWin, frmTime, windowX0, &
                           windowY0, windowDX, windowDY, binBPE)
    type(TableT), intent(in)         :: tableH
    integer(int32), intent(out) :: sciWin
    integer(int32), intent(out) :: frmTime
    integer(int32), intent(out) :: windowX0
    integer(int32), intent(out) :: windowY0
    integer(int32), intent(out) :: windowDX
    integer(int32), intent(out) :: windowDY
    logical, intent(out)             :: binBPE

    sciWin = int32Attribute(tableH, "SCIWIN")
    frmTime = int32Attribute(tableH, "FRMTIME")
    windowX0 = int32Attribute(tableH, "WINDOWX0")
    windowY0 = int32Attribute(tableH, "WINDOWY0")
    windowDX = int32Attribute(tableH, "WINDOWDX")
    windowDY = int32Attribute(tableH, "WINDOWDY")
    call GetBooleanTableAttribute(tableH, "BINBPE", binBPE)
    

end subroutine GetData2
!********************************************
! subroutine GetArrayData2
!********************************************
subroutine GetArrayData2(arrayH, sciWin, frmTime, windowX0, &
                           windowY0, windowDX, windowDY, &
                           binAx1, binAx2, binBPE)
    type(ArrayT), intent(in)         :: arrayH
    integer(int32), intent(out) :: sciWin
    integer(int32), intent(out) :: frmTime
    integer(int32), intent(out) :: windowX0
    integer(int32), intent(out) :: windowY0
    integer(int32), intent(out) :: windowDX
    integer(int32), intent(out) :: windowDY
    integer(int32), intent(out) :: binAx1
    integer(int32), intent(out) :: binAx2
    logical, intent(out)             :: binBPE


    sciWin = int32Attribute(arrayH, "SCIWIN")
    frmTime = int32Attribute(arrayH, "FRMTIME")
    windowX0 = int32Attribute(arrayH, "WINDOWX0")
    windowY0 = int32Attribute(arrayH, "WINDOWY0")
    windowDX = int32Attribute(arrayH, "WINDOWDX")
    windowDY = int32Attribute(arrayH, "WINDOWDY")

    binAx1 = GetInteger32ArrayAttribute(arrayH, "BINAX1")
    binAx2 = GetInteger32ArrayAttribute(arrayH, "BINAX2")
    call GetBooleanArrayAttribute(arrayH, "BINBPE", binBPE)
    

end subroutine GetArrayData2




!*************************************
! subroutine GetCommonTableData
!*************************************
subroutine GetCommonTableData(tableH, telescope, &
                         instrument, chain, dataType, &
                         obsId, expId, tmExpId, &
                         dateObs, dateEnd, exposure, binBPE)
    type(TableT), intent(in)         :: tableH
    character(*), intent(out)    :: telescope
    character(*), intent(out)    :: instrument
    integer(int32), intent(out) :: chain
    character(*), intent(out)    :: dataType
    character(*), intent(out)    :: obsId
    character(*), intent(out)    :: expId
    integer(int32), intent(out) :: tmExpId
    character(*), intent(out)    :: dateObs
    character(*), intent(out)    :: dateEnd
    real(single), intent(out)   :: exposure
    logical, intent(out)             :: binBPE
 
    telescope = stringAttribute(tableH, "TELESCOP")
    instrument = stringAttribute(tableH, "INSTRUME")
    chain = int32Attribute(tableH, "CHAIN")
    dataType = stringAttribute(tableH, "DATATYPE")
    obsId = stringAttribute(tableH, "OBS_ID")
    expId = stringAttribute(tableH, "EXP_ID")
    tmExpId = int32Attribute(tableH, "TMEXP_ID")
    dateObs = stringAttribute(tableH, "DATE-OBS")
    dateEnd = stringAttribute(tableH, "DATE-END")
    exposure = GetReal32TableAttribute(tableH, "EXPOSURE")
    call GetBooleanTableAttribute(tableH, "BINBPE", binBPE)
      
  
end subroutine GetCommonTableData
!*************************************
! subroutine GetCommonArrayData
!*************************************
subroutine GetCommonArrayData(arrayH, telescope, &
                         instrument, chain, dataType, &
                         obsId, expId, tmExpId, &
                         dateObs, dateEnd, exposure, binBPE)
    type(ArrayT), intent(in)         :: arrayH
    character(*), intent(out)    :: telescope
    character(*), intent(out)    :: instrument
    integer(int32), intent(out) :: chain
    character(*), intent(out)    :: dataType
    character(*), intent(out)    :: obsId
    character(*), intent(out)    :: expId
    integer(int32), intent(out) :: tmExpId
    character(*), intent(out)    :: dateObs
    character(*), intent(out)    :: dateEnd
    real(single), intent(out)   :: exposure
    logical, intent(out)             :: binBPE

    binBPE=.FALSE.
    telescope = stringAttribute(arrayH, "TELESCOP")
    instrument = stringAttribute(arrayH, "INSTRUME")
    chain = int32Attribute(arrayH, "CHAIN")
    dataType = stringAttribute(arrayH, "DATATYPE")
    obsId = stringAttribute(arrayH, "OBS_ID")
    expId = stringAttribute(arrayH, "EXP_ID")
    tmExpId = int32Attribute(arrayH, "TMEXP_ID")
    dateObs = stringAttribute(arrayH, "DATE-OBS")
    dateEnd = stringAttribute(arrayH, "DATE-END")
    exposure = GetReal32ArrayAttribute(arrayH, "EXPOSURE")
    call GetBooleanArrayAttribute(arrayH, "BINBPE", binBPE)

end subroutine GetCommonArrayData



!***************************************
! subroutine AddInt16Column1
!***************************************
subroutine AddInt16Column1(tableH, value, title) 
    type(tableT), intent(in) :: tableH
    integer(int16), dimension(:)    :: value   
    character(*), intent(in) :: title
    type(columnT) :: columnH
    integer(int16), dimension(:), pointer :: columnP   
 
    columnH = addColumn(tableH, title, Integer16)
    columnP => int16Data(columnH)
    columnP = value
    call release(columnH)
end subroutine AddInt16Column1
!***************************************
! subroutine AddInt16Column2
!***************************************
subroutine AddInt16Column2(tableH, value, title, units, columnDim) 
    type(tableT), intent(in) :: tableH
    integer(int16), intent(in), dimension(:) :: value
    character(*), intent(in) :: title
    character(*), intent(in) :: units
    integer(int32), intent(in), dimension(1) :: columnDim ! Column dimension.
    type(columnT) :: columnH
    integer(int16), dimension(:), pointer :: columnP   
 
    columnH = addColumn(tableH, title, Integer16, units, columnDim)
    columnP => int16Data(columnH)
    columnP = value
    call release(columnH)
end subroutine AddInt16Column2 
!***************************************
! subroutine AddInt32Column
!***************************************
subroutine AddInt32Column(tableH, value, title, units, columnDim) 
    type(tableT), intent(in) :: tableH
    integer(int32), intent(in) :: value
    character(*), intent(in) :: title
    character(*), intent(in) :: units
    integer(int32), intent(in), dimension(1) :: columnDim ! Column dimension.
    type(columnT) :: columnH
    integer(int32), dimension(:), pointer :: columnP   
 
    columnH = addColumn(tableH, title, Integer32, units, columnDim)
    columnP => int32Data(columnH)
    columnP = value
    call release(columnH)
end subroutine AddInt32Column 
!***************************************
! subroutine AddInt32Column1
!***************************************
subroutine AddInt32Column1(tableH, value, title) 
    type(tableT), intent(in) :: tableH
    integer(int32), dimension(:)    :: value   
    character(*), intent(in) :: title
    type(columnT) :: columnH
    integer(int32), dimension(:), pointer :: columnP   
 
    columnH = addColumn(tableH, title, Integer32)
    columnP => int32Data(columnH)
    columnP = value
    call release(columnH)
end subroutine AddInt32Column1 
!***************************************
! subroutine AddInt32Column1
!***************************************
subroutine AddInt32Column2(tableH, value, title) 

    type(tableT), intent(in) :: tableH
    integer(int32) :: value   
    character(len=*), intent(in) :: title
    type(columnT) :: columnH
    integer(int32), dimension(:), pointer :: columnP   
 
    columnH = addColumn(tableH, title, Integer32)
    columnP => int32Data(columnH)
    columnP = value
    call release(columnH)
end subroutine AddInt32Column2 
!***************************************
! subroutine AddInt32ArrayColumn
!***************************************
subroutine AddInt32ArrayColumn(tableH, value, title, units, columnDim) 

    type(tableT), intent(in) :: tableH
    integer(int32), dimension(:), pointer :: value             
    character(*), intent(in) :: title
    character(*), intent(in) :: units
    integer(int32), intent(in), dimension(1) :: columnDim ! Column dimension.
    type(columnT) :: columnH
    integer(int32), dimension(:), pointer :: columnP          

    columnH = addColumn(tableH, title, integer32, units, columnDim)
    columnP => int32Data(columnH)
    columnP = value
    call release(columnH)

end subroutine AddInt32ArrayColumn  
!***************************************
! subroutine AddInt32Array2Column
!***************************************
subroutine AddInt32Array2Column(tableH, value, title, units, columnDim) 

    type(tableT), intent(in) :: tableH
    integer(int32), dimension(:, :), pointer :: value             
    character(*), intent(in) :: title
    character(*), intent(in) :: units
    integer(int32), intent(in), dimension(1) :: columnDim ! Column dimension.
    type(columnT) :: columnH
    integer(int32), dimension(:, :), pointer :: columnP          

    columnH = addColumn(tableH, title, integer32, units, columnDim)
    columnP => int32VectorData(columnH)
    columnP = value
    call release(columnH)

end subroutine AddInt32Array2Column  
!************************************************
! subroutine CreateDummyTrackingHistoryFile
!************************************************
subroutine CreateDummyTrackingHistoryFile
    type(DataSetT) :: outDataSetH
    type(TableT)   :: outTableH

    outDataSetH = dataSet("tmp_tracking", create) 
    outTableH = addTable(outDataSetH, "ZERODRIFT", 1)
    call AddInt32Column2(outTableH, 0, "DRIFT") 
    call release(outTableH)
    call release(outDataSetH)

end subroutine CreateDummyTrackingHistoryFile
!****************************************
! CheckForNegativeCounts
!****************************************
subroutine CheckForNegativeCounts(image,  nx, ny) 
    integer, intent(in) :: nx, ny
    integer(int32), intent(inout), dimension(0 : nx, 0 : ny) :: image  
    integer i, j
    integer :: count
    count = 0

    do i = 0, nx
        do j = 0, ny
            if(image(i, j) .lt. 0) then
                image(i, j) = 0.0
                count = count + 1     
            end if
        end do
    end do

    if (count .gt. 0) then 
        write (errstr, *) count, " Negative Pixel values encountered"         
        call warning(trim(errstr), "PIXEL VALUES RESET TO ZERO")
    end if



end subroutine CheckForNegativeCounts

!*************************************
! Function GetReal32TableAttribute
!*************************************
real(kind=single) Function GetReal32TableAttribute(table, name)
   type(TableT), intent(in) :: table 
   character(*), intent(in) :: name
    
   if (hasAttribute(table, trim(name))) then
       GetReal32TableAttribute  = real32Attribute(table, trim(name))
   else
       GetReal32TableAttribute = -1.0

   end if
end Function GetReal32TableAttribute

!*************************************
! Function GetReal32ArrayAttribute
!*************************************
real(kind=single) Function GetReal32ArrayAttribute(array, name)
   type(ArrayT), intent(in) :: array 
   character(*), intent(in) :: name
    
   if (hasAttribute(array, trim(name))) then
       GetReal32ArrayAttribute  = real32Attribute(array, trim(name))
   else
       GetReal32ArrayAttribute = -1.0
   end if
end Function GetReal32ArrayAttribute

!*************************************
! Function GetReal32ArrayAttribute
!*************************************
integer(int32) Function GetInteger32ArrayAttribute(array, name)
   type(ArrayT), intent(in) :: array 
   character(len=*), intent(in) :: name
    
   if (hasAttribute(array, trim(name))) then
       GetInteger32ArrayAttribute  = real32Attribute(array, trim(name))
   else
       GetInteger32ArrayAttribute = -1.0
   end if
end Function GetInteger32ArrayAttribute

!*************************************
! subroutine GetReal32ArrayAttribute
!*************************************
subroutine  GetBooleanArrayAttribute(array, name, binBPE)
   type(ArrayT), intent(in) :: array 
   character(*), intent(in) :: name
   logical, intent(inout) :: binBPE
   
   binBPE=.FALSE. 
   if (hasAttribute(array, trim(name))) binBPE = booleanAttribute(array, trim(name))
   
end subroutine GetBooleanArrayAttribute
!*************************************
! GetReal32ArrayAttribute
!*************************************
subroutine GetBooleanTableAttribute(table, name, binBPE)
   type(TableT), intent(in) :: table 
   character(*), intent(in) :: name
   logical, intent(inout) :: binBPE  
   binBPE=.FALSE.
   if (hasAttribute(table, trim(name))) binBPE  = booleanAttribute(table, trim(name))
 
end subroutine GetBooleanTableAttribute


!*********************************************************
! Retrieves the exposure identifier from the given string
! ie S006, U4534, X008
!*********************************************************
function GetExposureIdentifier(name) result(exposureIdentifier)
    character(len=*), intent(in) :: name
    character(len=4) :: exposureIdentifier
    integer :: i, j, length
    character(len=4) :: string
    character(len=1) :: string1

    length = len(trim(name))

loop1 : do i = length - 7, 1, -1

        string = name(i : i + 3)
 
        if (string(1 : 1) .eq. 'S' .or. &
            string(1 : 1) .eq. 'U' .or. &
            string(1: 1) .eq. 'X') then
            
            do j = 2, 4
                string1 = string(j : j)
                if (scan(string1, "0123456789") .eq. 0) cycle loop1
               
            end do
 
            exposureIdentifier = string(1 : 4)

            if (exposureIdentifier(1 : 1) .eq.  "X") exposureIdentifier(1 : 1) = "S" 
            return
        end if
   
    end do loop1

    exposureIdentifier="NNNN" ! not found
    !call fatal("Exposure identifier not found")
    
end function GetExposureIdentifier





subroutine GetStartTime

    type(ExposureInfoType) :: expInfos
    type(SpacecraftAttitudeType) :: attitude  ! Attitude information
    
    call OAL_exposureInfo(expInfos)

end subroutine GetStartTime



end module omprep_mod









